use crate::Position;
use crate::Row;
use std::fs;
use std::io::{Error, Write};
use std::usize;

pub struct Document {
    rows: Vec<Row>,
    pub filename: Option<String>,
    pub dirty: bool,
    // auto save thread
}

impl Default for Document {
    fn default() -> Self {
        Self {
            rows: Vec::default(),
            filename: Option::default(),
            dirty: false,
        }
    }
}

impl Document {
    pub fn open(filename: &str) -> Result<Self, std::io::Error> {
        let contents = fs::read_to_string(filename)?;
        let mut rows = Vec::new();
        for value in contents.lines() {
            rows.push(Row::from(value))
        }
        Ok(Self {
            rows,
            filename: Some(String::from(filename)),
            dirty: false,
        })
    }

    pub fn filecheck(&self) -> bool {
        if let Some(filename) = &self.filename {
            fs::metadata(filename).is_ok()
        } else {
            false
        }
    }

    pub fn save(&mut self) -> Result<bool, Error> {
        if let Some(filename) = &self.filename {
            let mut file = fs::File::create(filename)?;
            for row in &self.rows {
                file.write_all(row.as_bytes())?;
                file.write_all(b"\n")?;
            }
            self.dirty = false;
            Ok(true)
        } else {
            Ok(false)
        }
    }

    pub fn row(&self, index: usize) -> Option<&Row> {
        self.rows.get(index)
    }

    pub fn is_empty(&self) -> bool {
        self.rows.is_empty()
    }

    pub fn len(&self) -> usize {
        self.rows.len()
    }

    pub fn insert_newline(&mut self, at: &Position) {
        if at.y > self.len() {
            return;
        }

        if at.y == self.len() {
            self.rows.push(Row::default());
        } else {
            let new_row = self.rows.get_mut(at.y).unwrap().split(at.x);
            self.rows.insert(at.y + 1, new_row)
        }
    }

    pub fn insert_text(&mut self, at: &Position, text: String) {
        let mut lines: Vec<Row> = text.split('\n').map(Row::from).collect();

        let prev_line = self.rows.get_mut(at.y).unwrap();
        let end_line = prev_line.split(at.x);

        let l = lines.len();

        if l == 1 {
            let new_row = Row::from(&text[..]);

            prev_line.append(&new_row);
            prev_line.append(&end_line);
        } else {
            prev_line.append(&lines[0]);

            lines[l - 1].append(&end_line);
            (1..lines.len()).for_each(|k| {
                let tmp = lines[k].clone();
                self.rows.insert(at.y + k, tmp);
            });
        }
    }

    // assumes from < to
    pub fn remove_text(&mut self, from: &Position, to: &Position) {
        let fromx = from.x;
        let fromy = from.y;

        let tox = to.x;
        let toy = to.y;

        let curline = self.rows.get_mut(fromy).unwrap();

        if fromy == toy {
            let mut rest = curline.split(fromx);

            curline.append(&rest.split(tox - fromx + 1));
        } else {
            curline.split(fromx);

            let lastline = self.rows.get_mut(toy).unwrap().split(tox + 1);
            self.rows[toy] = lastline;
            for k in fromy + 1..toy {
                self.rows.remove(fromy + 1);
            }
        }

        // self.rows.remove(index);
    }

    pub fn insert(&mut self, at: &Position, c: char) {
        // self.dirty = true;

        if c == '\n' {
            self.insert_newline(at);
            return;
        }

        if at.y == self.len() {
            let mut row = Row::default();
            row.insert(0, c);
            self.rows.push(row);
        } else if at.y < self.len() {
            let row = self.rows.get_mut(at.y).unwrap();
            row.insert(at.x, c);
        }
    }

    pub fn delete(&mut self, at: &Position) {
        let len = self.len();

        if at.y >= self.len() {
            return;
        }

        if at.x == self.rows.get_mut(at.y).unwrap().len() && at.y < len - 1 {
            let next_row = self.rows.remove(at.y + 1);
            let row = self.rows.get_mut(at.y).unwrap();
            row.append(&next_row);
        } else {
            let row = self.rows.get_mut(at.y).unwrap();
            row.delete(at.x);
        }
    }
}
