use crate::Document;
use crate::Events;
use crate::Row;
use crate::SharedInfo;
use crate::Statusbar;

use std::env;
use std::error::Error;
use std::sync::Arc;

use crossterm::event::{KeyCode, KeyEvent, KeyModifiers};
//use termion::event::Key;
use tokio::select;
use tokio::sync::mpsc::{UnboundedReceiver, UnboundedSender};
use tokio::sync::RwLock;

#[derive(Default)]
pub struct Position {
    pub x: usize,
    pub y: usize,
}

pub struct Editor {
    // terminal: Terminal,
    // pub document: Arc<RwLock<Document>>,
    // pub statusbar: Arc<RwLock<Statusbar>>,
    // pub cursor_position: Arc<RwLock<Position>>,
    pub shared: Arc<RwLock<SharedInfo>>,

    inputreceiver: UnboundedReceiver<KeyEvent>,
    eventsender: UnboundedSender<Events>,

    // offset: Position,
    rememberx: usize,
    quitting: bool,
    selectionstart: Option<Position>,
    selectionend: Option<Position>, // colorscheme: Colorscheme,
                                    // config: Config
}

impl Editor {
    pub fn default(irt: UnboundedReceiver<KeyEvent>, st: UnboundedSender<Events>) -> Self {
        let args: Vec<String> = env::args().collect();
        let document = if args.len() > 1 {
            let filename = &args[1];
            let mut d = Document::open(&filename).unwrap_or_default();
            d.filename = Some(filename.clone());
            d
        } else {
            Document::default()
        };

        // let terminal = Terminal::default().expect("Failed to initialize terminal");

        let shared = SharedInfo {
            cpos: Position::default(),
            stbar: Statusbar::new(st.clone()),
            doc: document,
        };

        Self {
            shared: Arc::new(RwLock::new(shared)),
            inputreceiver: irt,
            eventsender: st,

            // offset: Position::default(),
            rememberx: 0,
            quitting: false,
            
            selectionstart: None,
            selectionend: None
            // colorscheme: Colorscheme::default(),
            // config: Config::default(),
        }
    }

    fn refresh_screen(&self) -> Result<(), Box<dyn Error + Send + Sync>> {
        self.eventsender.send(Events::Refresh)?;

        Ok(())
    }

    fn refresh_status(&self) -> Result<(), Box<dyn Error + Send + Sync>> {
        self.eventsender.send(Events::RefreshStatus)?;

        Ok(())
    }

    fn quit(&self) -> Result<(), Box<dyn Error + Send + Sync>> {
        self.eventsender.send(Events::Quit)?;

        Ok(())
    }

    fn scroll(&self) -> Result<(), Box<dyn Error + Send + Sync>> {
        self.eventsender.send(Events::Scroll)?;

        Ok(())
    }

    pub async fn run(&mut self) -> Result<(), Box<dyn Error + Send + Sync>> {
        // let ticker = tick(Duration::from_millis(500));

        loop {
            match self.inputreceiver.recv().await {
                Some(k) => {
                    self.process_keypress(k).await?;
                    self.refresh_screen()?;
                }
                None => {}
            }
        }

        Ok(())
    }

    async fn process_keypress(&mut self, k: KeyEvent) -> Result<(), Box<dyn Error + Send + Sync>> {
        let sharedarc = self.shared.clone();
        let mut sharedlock = sharedarc.write().await;
        let shared = &mut *sharedlock;

        let mut doc = &mut shared.doc;
        let cpos = &mut shared.cpos;
        let stbar = &mut shared.stbar;

        match k {
            KeyEvent {
                code: KeyCode::Char('q'),
                modifiers: KeyModifiers::CONTROL,
            } => {
                if doc.dirty && !self.quitting {
                    stbar.change_status(
                        &"Unsaved changes. Are you sure you want to quit? Press ctrl+q again.",
                    )?;
                    self.quitting = true;
                } else {
                    self.quit()?;
                }

                return Ok(()); // Returing instantly, so that the self.quitting variable won't be reset to false
                               // At the end of this fucntion.
            }

            KeyEvent {
                code: KeyCode::Char('s'),
                modifiers: KeyModifiers::CONTROL,
            } => {
                if doc.dirty {
                    match doc.save() {
                        Ok(true) => {
                            stbar.change_status("Document saved successfully.")?;
                        }
                        Ok(false) => {
                            drop(sharedlock); // DROP THE LOCK BEFORE PROMPTING

                            let filename = self
                                .prompt(&String::from("Input filename"), &String::from(""))
                                .await?;

                            let sharedarc = self.shared.clone();
                            let shared = &mut *(*sharedarc).write().await;
                            let mut doc = &mut shared.doc;
                            let stbar = &mut shared.stbar;

                            doc.filename = filename;
                            if let Ok(true) = doc.save() {
                                stbar.change_status("Document saved successfully.")?;
                            } else {
                                stbar.change_status("Operation canceled.")?;
                            }
                        }
                        Err(e) => {
                            ();
                        } // TODO edit stuff
                    };
                } else {
                    stbar.change_status("No need to save, pal!")?;
                }
            }

            KeyEvent {
                code: KeyCode::Char('o'),
                modifiers: KeyModifiers::CONTROL,
            } => {
                drop(sharedlock); // DROP THE LOCK BEFORE PROMPTING

                let filename = self
                    .prompt(&String::from("Input filename"), &String::from(""))
                    .await?;

                let sharedarc = self.shared.clone();
                let shared = &mut *(*sharedarc).write().await;
                let mut doc = &mut shared.doc;
                let cpos = &mut shared.cpos;
                let stbar = &mut shared.stbar;

                match filename {
                    Some(fl) if !fl.is_empty() => {
                        *doc = Document::open(&fl).unwrap_or_default();
                        doc.filename = Some(fl);
                        stbar.change_status("Opening file")?;
                        *cpos = Position { x: 0, y: 0 };
                    }
                    _ => {
                        stbar.change_status("Operation canceled.")?;
                    }
                }
            }

            //            Key::Ctrl('t') => {
            //                doc.insert_text(cpos, String::from("WHEN THE \nMOON HITS YOU"));
            //                }
            // KeyEvent {
            //     code: KeyCode::Char('t'),
            //     modifiers: KeyModifiers::CONTROL,
            // } => {
            //     doc.remove_text(cpos, &Position { x: 2, y: 2 });
            // }
            //Key:: => {}
            KeyEvent {
                code: KeyCode::Char(c),
                ..
            } => {
                doc.dirty = true;
                doc.insert(cpos, c);
                self.move_cursor(
                    KeyEvent {
                        code: KeyCode::Right,
                        modifiers: KeyModifiers::NONE,
                    },
                    doc,
                    cpos,
                )
                .await?;
            }
            KeyEvent {
                code: KeyCode::Enter,
                ..
            } => {
                doc.dirty = true;
                doc.insert(cpos, '\n');
                self.move_cursor(
                    KeyEvent {
                        code: KeyCode::Right,
                        modifiers: KeyModifiers::NONE,
                    },
                    doc,
                    cpos,
                )
                .await?;
            }
            KeyEvent {
                code: KeyCode::Delete,
                ..
            } => {
                doc.dirty = true;
                doc.delete(cpos)
            }
            KeyEvent {
                code: KeyCode::Backspace,
                ..
            } => {
                if cpos.x > 0 || cpos.y > 0 {
                    doc.dirty = true;
                    self.move_cursor(
                        KeyEvent {
                            code: KeyCode::Left,
                            modifiers: KeyModifiers::NONE,
                        },
                        doc,
                        cpos,
                    )
                    .await?;
                    doc.delete(cpos);
                }
            }
            KeyEvent {
                code: KeyCode::Up, ..
            }
            | KeyEvent {
                code: KeyCode::Down,
                ..
            }
            | KeyEvent {
                code: KeyCode::Left,
                ..
            }
            | KeyEvent {
                code: KeyCode::Right,
                ..
            } => {
                self.move_cursor(k, doc, cpos).await?;
            }
            //            Key::Home => todo!(),
            //            Key::End => todo!(),
            //            Key::PageUp => todo!(),
            //            Key::PageDown => todo!(),
            //            Key::BackTab => todo!(),
            //            Key::Insert => todo!(),
            //            Key::F(_) => todo!(),
            //            Key::Alt(_) => todo!(),
            //            Key::Null => todo!(),
            //            Key::Esc => todo!(),
            //            Key::__IsNotComplete => todo!(),
            _ => {}
        }
        self.quitting = false;
        Ok(())
    }

    // fn savefile(&self) {
    //     doc.filecheck();
    // }

    async fn move_cursor(
        &mut self,
        key: KeyEvent,
        doc: &Document,
        cpos: &mut Position,
    ) -> Result<(), Box<dyn Error + Sync + Send>> {
        // let doc = (*docarc).read().await;
        // let mut cpos  = (*curarc).write().await;

        let Position { mut y, mut x } = *cpos;

        let w = doc.row(y).map_or(0, Row::len);
        let h = doc.len();

        let k = key.code;

        match k {
            KeyCode::Up => y = y.saturating_sub(1),
            KeyCode::Down => {
                if y < h {
                    y = y.saturating_add(1)
                }
            }
            KeyCode::Left => {
                if x > 0 {
                    x -= 1;
                } else if y > 0 {
                    y -= 1;
                    x = doc.row(y).map_or(0, Row::len);
                }
                self.rememberx = x;
            }
            KeyCode::Right => {
                if x < w {
                    x += 1
                } else if y < h {
                    y += 1;
                    x = 0;
                }
                self.rememberx = x;
            }
            _ => (),
        }
        let neww = doc.row(y).map_or(0, Row::len);

        if neww <= self.rememberx {
            x = neww;
        } else {
            x = self.rememberx
        };

        *cpos = Position { x, y };
        self.scroll()
    }

    async fn prompt(
        &mut self,
        prompt_message: &String,
        original: &String,
    ) -> Result<Option<String>, Box<dyn Error + Send + Sync>> {
        let sharedarc = self.shared.clone();
        let mut sharedlock = sharedarc.write().await;
        let shared = &mut *sharedlock;

        let stbar = &mut shared.stbar;

        stbar.prompting = true;
        stbar.status_stat = prompt_message.clone();

        let mut pos = original.len();

        let mut r = Row::from(&original[..]);

        stbar.status2_stat = r.render(0, r.len());

        drop(sharedlock); // DROP BEFORE REFRESHING

        self.refresh_status()?;

        loop {
            select!(
                k = self.inputreceiver.recv() => {
                    {
                        let mut sharedlock = sharedarc.write().await;
                        let shared = &mut *sharedlock;
                        let stbar = &mut shared.stbar;

                        match k.unwrap() {

                            KeyEvent{ code: KeyCode::Left, modifiers : KeyModifiers::NONE} => {
                                if pos > 0 {
                                    pos = pos - 1
                                }
                            }

                            KeyEvent{ code: KeyCode::Right, modifiers : KeyModifiers::NONE} => {
                                if pos < r.len() {
                                    pos = pos + 1
                                }

                            }

                            KeyEvent{ code: KeyCode::Enter, modifiers : KeyModifiers::NONE} => {
                                stbar.prompting = false;
                                if r.is_empty() {
                                    return Ok(None)
                                } else {
                                    return Ok(
                                        Some(r.render(0, r.len()))
                                    )
                                }
                            }


                            KeyEvent{ code: KeyCode::Esc, modifiers : KeyModifiers::NONE} => {
                                stbar.prompting = false;
                                return Ok(None)
                            }

                            KeyEvent{ code: KeyCode::Backspace, modifiers : KeyModifiers::NONE} => {
                                if pos > 0 {
                                    r.delete(pos-1);
                                    pos = pos - 1
                                }
                            }

                            KeyEvent{ code: KeyCode::Char(c), modifiers : KeyModifiers::NONE} => {
                                    r.insert(pos, c);
                                    pos = pos + 1
                                }

                            _ => {}
                        }
                        stbar.status2_stat = r.render(0, r.len());
                    }
                    self.refresh_status()?; // everything is dropped automatically, refresh again after each character
            })
        }
    }
}
