use crate::Colorscheme;
use crate::Config;
use crate::Events;
use crate::Position;
use crate::Row;
use crate::SharedInfo;
use crate::Terminal;

use std::cmp;
use std::error::Error;
use std::sync::Arc;

use crossterm::event::KeyEvent;

use tokio::select;
use tokio::sync::mpsc::{UnboundedReceiver, UnboundedSender};
use tokio::sync::RwLock;

pub struct Renderer {
    shared: Arc<RwLock<SharedInfo>>,
    terminal: Terminal,

    // render_position: Position,
    offset: Position,
    colorscheme: Colorscheme,

    pub receiver: UnboundedReceiver<Events>,
    config: Config,
}

impl Renderer {
    pub fn default(
        st: UnboundedSender<Events>,
        ist: UnboundedSender<KeyEvent>,
        rt: UnboundedReceiver<Events>,
        shared: Arc<RwLock<SharedInfo>>,
    ) -> Self {
        // let (st, rt) = unbounded_channel();
        // let (ist, irt) = unbounded_channel();

        let terminal = Terminal::default(st.clone(), ist).expect("Failed to initialize terminal");

        // let editor = Editor::default(irt, st);

        Self {
            shared,
            terminal,

            receiver: rt,

            // render_position: Position::default(),
            offset: Position::default(),
            colorscheme: Colorscheme::default(),

            config: Config::default(),
        }
    }

    pub async fn run(&mut self) -> Result<(), Box<dyn Error + Send + Sync>> {
        self.terminal.enable_alt_screen();

        self.refresh_screen().await?;

        loop {
            select! {
                event = self.receiver.recv() => {match event.unwrap() {
                    Events::Quit => {
                        break;
                    },
                    Events::Refresh => {self.refresh_screen().await?;}
                    Events::RefreshStatus => {self.refresh_status().await?;}
                    Events::Scroll => {self.scroll().await;}
                }}

            }
        }

        self.terminal.disable_alt_screen();
        Ok(())
    }

    async fn calculate_pos(&self) -> Position {
        let sharedarc = self.shared.clone();

        let shared = (*sharedarc).read().await;

        let doc = &shared.doc;
        let cpos = &shared.cpos;

        let Position { x, y } = *cpos;
        let Position { x: ox, y: oy } = self.offset;

        let w = self.terminal.size().width as usize;

        let maxw = self.config.maxwidth;

        let wpad = (w.saturating_sub(maxw)) / 2;
        let ypad = self.config.vertical_padding;

        let mut linewidth = 0;

        if let Some(r) = doc.row(y) {
            linewidth = r.width(ox, x);
        }
        Position {
            x: linewidth + wpad,
            y: (y + ypad).saturating_sub(oy),
        }
    }

    async fn refresh_screen(&mut self) -> Result<(), Box<dyn Error + Send + Sync>> {
        self.terminal.cursor_hide()?;
        self.draw_rows().await;
        self.draw_info().await;

        self.terminal.cursor_position(&self.calculate_pos().await);
        self.terminal.cursor_show()?;
        self.terminal.flush()?;
        Ok(())
    }

    async fn refresh_status(&mut self) -> Result<(), Box<dyn Error + Send + Sync>> {
        self.terminal.cursor_hide()?;

        let s = self.terminal.size();

        // let w = s.width as usize;
        let h = s.height as usize - 2;

        self.terminal.cursor_position(&Position { x: 0, y: h });

        self.draw_info().await;

        self.terminal.cursor_position(&self.calculate_pos().await);
        self.terminal.cursor_show();
        self.terminal.flush()?;
        Ok(())
    }

    fn draw_row(&self, row: &Row) {
        let w = self.terminal.size().width as usize;

        let maxw = self.config.maxwidth;

        let pad = (w.saturating_sub(maxw)) / 2;

        let start = self.offset.x;
        let end = self.offset.x + w - pad * 2;
        let r = row.render(start, end);
        let space = &" ".repeat(pad);
        println!("{}{}\r", space, r)
    }

    async fn draw_rows(&mut self) {
        let sharedarc = self.shared.clone();

        let shared = (*sharedarc).read().await;

        let doc = &shared.doc;

        self.terminal.cursor_reset();
        self.terminal.set_text_color(self.colorscheme.text);
        self.terminal.set_bg_color(self.colorscheme.background);

        let h = self.terminal.size().height as usize;
        // let w = self.terminal.size().width as usize;

        let vertical_pad = self.config.vertical_padding.saturating_sub(self.offset.y);

        let trueh = h.saturating_sub(vertical_pad);

        for _ in 0..vertical_pad {
            self.terminal.clear_current_line();
            println!("\r")
        }

        for terminal_row in 0..trueh - 2 {
            self.terminal.clear_current_line();
            if let Some(row) = doc.row(
                terminal_row as usize + self.offset.y.saturating_sub(self.config.vertical_padding),
            ) {
                self.draw_row(row);
            } else {
                println!("\r");
            }
        }
        // Terminal::reset_color();
    }

    async fn draw_info(&mut self) {
        let sharedarc = self.shared.clone();

        let shared = (*sharedarc).read().await;

        let doc = &shared.doc;
        let cpos = &shared.cpos;
        let stbar = &shared.stbar;

        self.terminal
            .set_bg_color(self.colorscheme.status_background);
        self.terminal.set_text_color(self.colorscheme.status_text);

        let width = self.terminal.size().width as usize;

        let status = stbar.status();

        let status2 = stbar.status2(&doc, &cpos, width);

        self.terminal.clear_current_line();
        println!("{}\r", status);
        self.terminal.clear_current_line();
        print!("{}\r", status2);

        self.terminal.reset_color();
    }

    async fn scroll(&mut self) {
        let sharedarc = self.shared.clone();

        let shared = (*sharedarc).read().await;

        let cpos = &shared.cpos;

        let Position { x, y } = *cpos;
        let s = self.terminal.size();
        let w = s.width as usize;
        let h = s.height as usize - 2;

        // let yscroll = self.config.vertical_scrollspace;
        // let xscroll = self.config.horizontal_scrollspace;

        let ypad = self.config.vertical_padding;
        let maxw = self.config.maxwidth;

        let mut offset = &mut self.offset;

        let true_w = cmp::min(w, maxw);

        if y - offset.y.saturating_sub(ypad) + 1 > h - ypad.saturating_sub(offset.y) {
            offset.y = y.saturating_sub(h - ypad).saturating_add(1);
        } else if y < offset.y {
            offset.y = y;
        }

        if x + 1 > offset.x.saturating_add(true_w) {
            offset.x = x.saturating_sub(true_w).saturating_add(1);
        } else if x < offset.x {
            offset.x = x;
        }
    }
}
