use crate::Events;
use crate::Position;

use std::cmp;
use std::error::Error;
use std::io::{stdout, Stdout, Write};
use std::sync::{Arc, RwLock};
use std::thread;
use std::time::Duration;

use crossterm::{
    cursor::{Hide, MoveTo, Show},
    event::{poll, read, Event, KeyEvent},
    execute, queue,
    style::{Color, ResetColor, SetBackgroundColor, SetForegroundColor},
    terminal::{
        enable_raw_mode, size, Clear, ClearType, EnterAlternateScreen, LeaveAlternateScreen,
    },
};

use tokio::signal::unix::{signal, SignalKind};
use tokio::sync::mpsc::UnboundedSender;

/// Structure to represent the size of the terminal.
#[derive(Clone, Copy)]
pub struct Size {
    pub width: u16,
    pub height: u16,
}

/// Structure to represent a terminal interface.
///
/// Size is stored in a arclock, so it can be updated from another thread.
pub struct Terminal {
    size: Arc<RwLock<Size>>,

    stdout: Stdout,

    pub sender: UnboundedSender<Events>, // sends screen refresh events to the renderer.
    inputsender: UnboundedSender<KeyEvent>, // sends input to the editor

    keythread: Option<std::thread::JoinHandle<()>>, // thread that reads input
}

impl Terminal {
    pub fn default(
        st: UnboundedSender<Events>,
        ist: UnboundedSender<KeyEvent>,
    ) -> Result<Self, Box<dyn Error + Send + Sync>> {
        let size = size()?;
        enable_raw_mode()?;

        let stdout = stdout();

        let mut res = Self {
            size: Arc::new(RwLock::new(Size {
                width: size.0,
                height: size.1,
            })),
            stdout,
            sender: st,
            inputsender: ist,
            keythread: None,
        };

        res.start_reading()?;

        Ok(res)
    }

    pub fn enable_alt_screen(&mut self) -> Result<(), Box<dyn Error + Send + Sync>> {
        execute!(self.stdout, EnterAlternateScreen)?;
        Ok(())
    }

    pub fn disable_alt_screen(&mut self) -> Result<(), Box<dyn Error + Send + Sync>> {
        execute!(self.stdout, LeaveAlternateScreen)?;
        Ok(())
    }
    fn start_reading(&mut self) -> Result<(), Box<dyn Error + Send + Sync>> {
        //let mut reader = stdin().keys();

        let isender = self.inputsender.clone();
        let sender = self.sender.clone();

        let sarc = self.size.clone();

        let keythread = thread::spawn(move || loop {
            loop {
                if poll(Duration::from_millis(500)).unwrap() {
                    // It's guaranteed that the `read()` won't block when the `poll()`
                    // function returns `true`
                    match read().unwrap() {
                        Event::Key(event) => {
                            isender.send(event);
                        }
                        Event::Mouse(event) => {}
                        // Mouse events will be supported later?
                        Event::Resize(width, height) => {
                            let mut w = sarc.write().unwrap();
                            *w = Size { width, height };
                            sender.send(Events::Refresh);
                        }
                    }
                } else {
                    sender.send(Events::RefreshStatus);
                }
            }
        });

        self.keythread = Some(keythread);
        Ok(())
    }

    pub fn size(&self) -> Size {
        let w = self.size.read().unwrap();
        *w
    }

    pub fn clear_screen(&mut self) -> Result<(), Box<dyn Error + Send + Sync>> {
        queue!(self.stdout, Clear(ClearType::All))?;
        Ok(())
    }

    pub fn clear_current_line(&mut self) -> Result<(), Box<dyn Error + Send + Sync>> {
        queue!(self.stdout, Clear(ClearType::CurrentLine))?;
        Ok(())
    }

    pub fn cursor_reset(&mut self) -> Result<(), Box<dyn Error + Send + Sync>> {
        queue!(self.stdout, MoveTo(1, 1))?;
        Ok(())
    }

    pub fn cursor_position(&mut self, p: &Position) {
        let Position { x, y } = p;
        let s = self.size();
        let x = cmp::min(*x as u16, s.width);
        let y = cmp::min(*y as u16, s.height);
        queue!((self.stdout), MoveTo(x, y));
    }

    pub fn flush(&mut self) -> Result<(), Box<dyn Error + Send + Sync>> {
        self.stdout.flush()?;
        Ok(())
    }

    pub fn cursor_hide(&mut self) -> Result<(), Box<dyn Error + Send + Sync>> {
        queue!(self.stdout, Hide)?;
        Ok(())
    }

    pub fn cursor_show(&mut self) -> Result<(), Box<dyn Error + Send + Sync>> {
        queue!(self.stdout, Show)?;
        Ok(())
    }

    pub fn set_bg_color(&mut self, color: Color) -> Result<(), Box<dyn Error + Send + Sync>> {
        queue!(self.stdout, SetBackgroundColor(color))?;
        Ok(())
    }

    pub fn set_text_color(&mut self, color: Color) -> Result<(), Box<dyn Error + Send + Sync>> {
        queue!(self.stdout, SetForegroundColor(color))?;
        Ok(())
    }

    pub fn reset_color(&mut self) -> Result<(), Box<dyn Error + Send + Sync>> {
        queue!(self.stdout, ResetColor)?;
        Ok(())
    }

    pub fn quit(&self) -> Result<(), Box<dyn Error>> {
        self.sender.send(Events::Quit)?;
        Ok(())
    }
}
